from django.template.loader import render_to_string
from django.contrib.auth.models import User
from django.core.mail import send_mail
from django.conf import settings

from .models import FailedLogin

import datetime


def email_user(username, password, **kwargs):
    try:
        user = User.objects.get(username=username)
        message = render_to_string("secure_login/failed_login_user.txt")
        send_mail("failed_login",
                  message,
                  settings.DEFAULT_FROM_EMAIL,
                  [user.email])

    except User.DoesNotExist:
        pass


def populate_failed_requests(username, password, **kwargs):
    request = kwargs.get("request")
    if request and get_client_ip(request):
        ip = get_client_ip(request)
    else:
        ip = None
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        user = None
    FailedLogin.objects.create(for_user=user, ip=ip)


def lockout_on_many_wrong_password(username, password, **kwargs):
    try:
        user = User.objects.get(username=username)
        one_hour_ago = datetime.datetime.now() - datetime.timedelta(hours=1)
        failed_logins = FailedLogin.objects.filter(for_user=user, attempted_at__gte=one_hour_ago).count()
        max_hourly_attempts = getattr(settings, "SECURE_LOGIN_MAX_HOURLY_ATTEMPTS", 10)
        if failed_logins >= max_hourly_attempts:
            user.is_active = False
            user.save()
    except User.DoesNotExist:
        pass


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    ip = None
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip



from django.conf import settings

import os


BASE_DIR = os.path.abspath(os.path.dirname(__file__))

wordlist = os.path.join(BASE_DIR, "weakpasswords.txt")
weak_passwords = [el.strip() for el in open(wordlist).read().split()]


def no_weak_passwords(username, password, **kwargs):
    if password in weak_passwords:
        return False
    return True


def no_short_passwords(username, password, **kwargs):
    if (len(password) <
            getattr(settings, "SECURE_LOGIN_MIN_PASSWORD_LENGTH", 6)):
        return False
    return True


def no_username_password_same(username, password, **kwargs):
    if username == password:
        return False
    return True



from django.db import models

from django.contrib.auth.models import User


class FailedLogin(models.Model):
    attempted_at = models.DateTimeField(auto_now=True)
    for_user = models.ForeignKey(User, null=True, blank=True)
    ip = models.GenericIPAddressField(null=True, blank=True)



from django.conf import settings
from django.contrib.auth.forms import AuthenticationForm
from django import forms

from .utils import get_callable, handle_fieldname

DEFAULT_ERROR_MESSAGE = "Please review the username and password"


class SecureFormMixin(object):

    def clean(self):
        DEFAULT_CHECKERS = ["secure_login.checkers.no_weak_passwords",
                            "secure_login.checkers.no_short_passwords",
                            "secure_login.checkers.no_username_password_same"]

        checkers = getattr(settings, "SECURE_LOGIN_CHECKERS", DEFAULT_CHECKERS)
        for checker in checkers:
            checker_ = get_callable(checker)
            checker_ = handle_fieldname(self.username_fieldname(), self.password_fieldname(), checker_)
            if not checker_(**self.cleaned_data):
                raise forms.ValidationError(getattr(checker_, "error_message", DEFAULT_ERROR_MESSAGE))
        return super(SecureFormMixin, self).clean()

    def username_fieldname(self):
        return "username"

    def password_fieldname(self):
        return "password"


class SecureLoginForm(SecureFormMixin, AuthenticationForm):
    pass



from django.contrib.auth import backends
from django.conf import settings


from .utils import get_callable, handle_fieldname


class SecureLoginBackendMixin(object):

    def authenticate(self, **kwargs):
        DEFAULT_CHECKERS = ["secure_login.checkers.no_weak_passwords",
                            "secure_login.checkers.no_short_passwords",
                            "secure_login.checkers.no_username_password_same"]

        checkers = getattr(settings, "SECURE_LOGIN_CHECKERS", DEFAULT_CHECKERS)
        request = kwargs.pop('request', None)

        DEFAULT_ON_FAIL = ["secure_login.on_fail.email_user", ]
        on_fail_callables = getattr(settings,
                                    "SECURE_LOGIN_ON_FAIL",
                                    DEFAULT_ON_FAIL)

        checker_failed = False
        for checker in checkers:
            checker_ = self.get_final_callable(checker)
            if not checker_(**kwargs):
                checker_failed = True
                break
        if checker_failed:
            for callable_ in on_fail_callables:
                self.get_final_callable(callable_)(**kwargs)
            return None

        user = super(SecureLoginBackendMixin, self).authenticate(**kwargs)

        if not user:  # Login failed
            for callable_ in on_fail_callables:
                callable_ = self.get_final_callable(callable_)
                callable_(**kwargs)
        return user

    def username_fieldname(self):
        return "username"

    def password_fieldname(self):
        return "password"

    def get_final_callable(self, checker):
        checker_ = get_callable(checker)
        return handle_fieldname(self.username_fieldname(), self.password_fieldname(), checker_)


class SecureLoginBackend(SecureLoginBackendMixin, backends.ModelBackend):
    pass



from django.conf import settings


def get_callable(callable_str):
    path = callable_str.split(".")
    module_name = ".".join(path[:-1])
    callable_name = path[-1]
    module = __import__(module_name, {}, {}, [callable_name])
    callable_ = getattr(module, callable_name)
    return callable_


def handle_fieldname(username_field, password_field, func):
    def deco(**kwargs):
        username = kwargs.get(username_field)
        password = kwargs.get(password_field)
        if username:
            del kwargs[username_field]
        if password:
            del kwargs[password_field]
        return func(username, password, **kwargs)
    return deco



from django.test import TestCase
from django.contrib.auth.models import User
from django.contrib.auth import authenticate
from django.core import mail
from django.test.utils import override_settings
from django.conf import settings
from django import forms
from django.contrib.auth.backends import ModelBackend

from .models import FailedLogin
from .forms import SecureLoginForm, SecureFormMixin
from .backends import SecureLoginBackendMixin


class SecureLoginBackendTest(TestCase):

    @override_settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_weak_passwords", ])
    def test_no_weak_passwords(self):
        bad_password = "albatross"
        good_password = "a-l0ng-pa55w0rd-@^&"
        user = User.objects.create(username="hello")
        user.set_password(bad_password)
        user.save()
        self.assertFalse(authenticate(username="hello", password=bad_password))

        user.set_password(good_password)
        user.save()
        self.assertEqual(
            authenticate(username="hello", password=good_password), user)

    @override_settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_short_passwords", ])
    def test_no_short_passwords(self):
        bad_password = "123"
        good_password = "a-l0ng-pa55w0rd-@^&"
        user = User.objects.create(username="hello")
        user.set_password(bad_password)
        user.save()
        self.assertFalse(authenticate(username="hello", password=bad_password))

        user.set_password(good_password)
        user.save()
        self.assertEqual(
            authenticate(username="hello", password=good_password), user)

    @override_settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_username_password_same", ])
    def test_no_username_password_same(self):
        username = "hellohello"
        bad_password = "hellohello"
        good_password = "a-l0ng-pa55w0rd-@^&"
        user = User.objects.create(username=username)
        user.set_password(bad_password)
        user.save()
        self.assertFalse(
            authenticate(username=username, password=bad_password))

        user.set_password(good_password)
        user.save()
        self.assertEqual(
            authenticate(username=username, password=good_password), user)

    @override_settings(SECURE_LOGIN_ON_FAIL=["secure_login.on_fail.email_user", ])
    def test_email_sent_on_wrong_password(self):
        username = "hello"
        password = "hellohello"
        user = User.objects.create(username=username)
        user.set_password(password)
        user.save()

        self.assertFalse(
            authenticate(username=username, password=password + "1"))
        self.assertEqual(len(mail.outbox), 1)

    @override_settings(SECURE_LOGIN_ON_FAIL=["secure_login.on_fail.populate_failed_requests"], SECURE_LOGIN_CHECKERS=[])
    def test_populate_failed_requests(self):
        username = "hello"
        password = "hellohello"
        user = User.objects.create_user(username=username, password=password)

        authenticate(username=username, password="not-the-correct-password")
        self.assertEqual(FailedLogin.objects.count(), 1)

    @override_settings(SECURE_LOGIN_ON_FAIL=["secure_login.on_fail.populate_failed_requests", "secure_login.on_fail.lockout_on_many_wrong_password", ], SECURE_LOGIN_CHECKERS=[])
    def test_lockout(self):
        username = "hello"
        password = "hellohello"
        user = User.objects.create_user(username=username, password=password)

        for _ in range(11):
            authenticate(
                username=username, password="not-the-correct-password")
        user_ = authenticate(username=username, password=password)
        self.assertFalse(user_.is_active)

    def test_email_based_backend(self):

        username = "hello"
        password = "albatross"
        email = "hello@example.com"
        user = User.objects.create_user(username=username, password=password, email=email)

        with self.settings(AUTHENTICATION_BACKENDS=["secure_login.tests.SecureEmailBackend"], SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_weak_passwords"]):
            self.assertEqual(authenticate(email=email, password=password), None)

        with self.settings(AUTHENTICATION_BACKENDS=["secure_login.tests.SecureEmailBackend"], SECURE_LOGIN_CHECKERS=[]):
            self.assertEqual(authenticate(email=email, password=password), user)

    def test_multiple_backend(self):
        username = "hello"
        password = "albatross"
        email = "hello@example.com"
        user = User.objects.create_user(username=username, password=password, email=email)

        with self.settings(AUTHENTICATION_BACKENDS=["secure_login.tests.SecureEmailBackend", "secure_login.tests.SecureUsernameBackend"], SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_weak_passwords"]):
            self.assertEqual(authenticate(email=email, password=password), None)


class FormsTest(TestCase):

    @override_settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_weak_passwords", ])
    def test_no_weak_passwords(self):
        bad_password = "albatross"
        good_password = "a-l0ng-pa55w0rd-@^&"

        user = User.objects.create(username="hello")
        user.set_password(bad_password)
        user.save()

        form = SecureLoginForm(
            data={"username": "hello", "password": bad_password})
        self.assertFalse(form.is_valid())

        user.set_password(good_password)
        user.save()
        form = SecureLoginForm(
            data={"username": "hello", "password": good_password})
        self.assertTrue(form.is_valid())

    @override_settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_short_passwords", ])
    def test_no_short_passwords(self):
        bad_password = "123"
        good_password = "a-l0ng-pa55w0rd-@^&"
        user = User.objects.create(username="hello")
        user.set_password(bad_password)
        user.save()
        form = SecureLoginForm(
            data={"username": "hello", "password": bad_password})

        self.assertFalse(form.is_valid())

        user.set_password(good_password)
        user.save()
        form = SecureLoginForm(
            data={"username": "hello", "password": good_password})
        self.assertTrue(form.is_valid())

    def test_register_form(self):
        class RegiterForm(forms.Form):
            username = forms.CharField(max_length=50)
            password = forms.CharField()
            email = forms.EmailField(required=False)

        class SecureRegisterForm(SecureFormMixin, RegiterForm):
            pass

        bad_password = "albatross"

        with self.settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_weak_passwords", ]):
            form = SecureRegisterForm(
                data={"username": "hello", "password": bad_password})
            self.assertFalse(form.is_valid())

        with self.settings(SECURE_LOGIN_CHECKERS=[]):
            form = SecureRegisterForm(
                data={"username": "hello", "password": bad_password})
            self.assertTrue(form.is_valid())

        bad_password = 123
        with self.settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_short_passwords", ]):
            form = SecureRegisterForm(
                data={"username": "hello", "password": bad_password})
            self.assertFalse(form.is_valid())

        with self.settings(SECURE_LOGIN_CHECKERS=[]):
            form = SecureRegisterForm(
                data={"username": "hello", "password": bad_password})
            self.assertTrue(form.is_valid())

    def test_email_login_form(self):
        class EmailLoginForm(forms.Form):
            email = forms.EmailField()
            password = forms.CharField()

        class SecureRegisterForm(SecureFormMixin, EmailLoginForm):
            pass

            def username_fieldname(self):
                return "email"

        bad_password = "albatross"

        with self.settings(SECURE_LOGIN_CHECKERS=["secure_login.checkers.no_weak_passwords", ]):
            form = SecureRegisterForm(
                data={"email": "hello@example.com", "password": bad_password})
            self.assertFalse(form.is_valid())

        with self.settings(SECURE_LOGIN_CHECKERS=[]):
            form = SecureRegisterForm(
                data={"email": "hello@example.com", "password": bad_password})
            self.assertTrue(form.is_valid())


class EmailBackend(object):
    def authenticate(self, email, password, **kwargs):
        try:
            return User.objects.get(email=email)
        except User.DoesNotexist:
            return None


class SecureEmailBackend(SecureLoginBackendMixin, EmailBackend):
    def username_fieldname(self):
        return "email"


class SecureUsernameBackend(SecureLoginBackendMixin, ModelBackend):
    pass
